/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*-

   this file is part of rcssserver3D
   Fri May 9 2003
   Copyright (C) 2002,2003 Koblenz University
   Copyright (C) 2003 RoboCup Soccer Server 3D Maintenance Group
   $Id: worldmodel.cpp,v 1.4 2004/07/01 18:31:48 kubo Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "worldmodel.h"

using namespace zeitgeist;
using namespace std;
using namespace boost;
using namespace oxygen;
using namespace salt;

WorldModel::WorldModel()
{
    mFieldLength = 105.0;
    mFieldWidth = 68.0;
    mFieldHeight = 20.0;
    mGoalWidth = 7.32;
    mGoalDepth = 2.0;
    mGoalHeight = 2.44;
    mBorderSize = 4.0f;

    // agent parameter
    mAgentMass = 75;
    mAgentRadius = 0.22;
    mAgentMaxSpeed = 10;

    // ball parameter
    mBallRadius = 0.111;
    mBallMass = 0.4;

    // team information
    mTeamIndex = TI_NONE;
    mTeamUnum = 0;


    mMyPosVEnable = false;

    Formation* fp = &(formations[FT_initial]);
    int i = 1;
    fp->player[i++].setValues(-50.0,	0.0,	0.0,	0.0,	-49.0,	0.0, false);
    fp->player[i++].setValues(-16.0,	16.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-17.0,	5.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-17.0,	-5.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-16.0,	-16.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-8.0,	0.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-5.0,	10.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-5.0,	-10.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-2.0,	0.0,	0.0,	0.0,	-45.0,	0.0, false);
    fp->player[i++].setValues(-1.0,	22.0);
    fp->player[i++].setValues(-1.0,	-22.0);

    fp = &(formations[FT_433_offensive]);
    i = 1;
    fp->player[i++].setValues(-50.0,	0.0,	0.0,	0.0,	-52.0,	-50.0,	false);
    fp->player[i++].setValues(-16.5,	10.0,	0.1,	0.1,	-50.5,	-30.0,	true);
    fp->player[i++].setValues(-21.0,	0.0,	0.7,	0.2,	-42.0,	0.0,	true);
    fp->player[i++].setValues(-15.0,	0.0,	0.65,	0.4,	-47.0,	2.0,	true);
    fp->player[i++].setValues(-16.5,	-10.0,	0.7,	0.25,	-45.0,	2.0,	false);
    fp->player[i++].setValues(0.0,	-11.0,	0.65,	0.3,	-36.0,	42.0,	false);
    fp->player[i++].setValues(0.0,	11.0,	0.7,	0.25,	-36.0,	42.0,	false);
    fp->player[i++].setValues(-3.0,	0.5,	0.5,	0.3,	-2.0,	44.0,	false);
    fp->player[i++].setValues(15.0,	-0.5,	0.6,	0.25,	-2.0,	44.0,	false);
    fp->player[i++].setValues(18.0,	19.0);
    fp->player[i++].setValues(18.0,	-19.0);

    fp = &(formations[FT_334_offensive]);
    i = 1;
    fp->player[i++].setValues(-50.0,	0.0,	0.0,	0.0,	0.0,	0.0,	false);
    fp->player[i++].setValues(-16.0,	10.0,	0.1,	0.1,	-50.5,	-30.0,	true);
    fp->player[i++].setValues(-21.0,	0.0,	0.6,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-5.0,	0.5,	0.6,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-16.0,	-10.0,	0.6,	0.25,	-45.0,	2.0,	false);
    fp->player[i++].setValues(3.0,	-11.0,	0.6,	0.25,	-30.0,	20.0,	false);
    fp->player[i++].setValues(3.0,	11.0,	0.6,	0.3,	-30.0,	42.0,	false);
    fp->player[i++].setValues(15.0,	-5.0,	0.5,	0.2,	-45.0,	44.0,	false);
    fp->player[i++].setValues(15.0,	5.0,	0.6,	0.2,	-2.0,	45.0,	false);
    fp->player[i++].setValues(18.0,	19.0);
    fp->player[i++].setValues(18.0,	-19.0);

    fp = &(formations[FT_defensive]);
    i = 1;
    fp->player[i++].setValues(-45.0,	0.0,	0.0,	0.0,	0.0,	0.0,	false);
    fp->player[i++].setValues(-12.0,	21.0,	0.1,	0.1,	-52.5,	-30.0,	true);
    fp->player[i++].setValues(-14.0,	5.0,	0.5,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-14.0,	-5.0,	0.5,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-12.0,	-21.0,	0.4,	0.2,	-45.0,	35.0,	false);
    fp->player[i++].setValues(1.0,	24.0,	0.6,	0.3,	-30.0,	42.0,	false);
    fp->player[i++].setValues(-5.0,	8.0,	0.5,	0.15,	-30.0,	42.0,	false);
    fp->player[i++].setValues(-5.0,	-8.0,	0.5,	0.25,	-45.0,	47.0,	false);
    fp->player[i++].setValues(1.0,	-24.0,	0.6,	0.3,	10.0,	47.0,	false);
    fp->player[i++].setValues(18.0,	7.0);
    fp->player[i++].setValues(18.0,	-7.0);

    fp = &(formations[FT_open_defensive]);
    i = 1;
    fp->player[i++].setValues(-45.0,	0.0,	0.0,	0.0,	0.0,	0.0,	false);
    fp->player[i++].setValues(-11.0,	18.0,	0.1,	0.1,	-52.5,	-30.0,	true);
    fp->player[i++].setValues(-12.0,	7.0,	0.4,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-12.0,	-7.0,	0.4,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-11.0,	-18.0,	0.6,	0.25,	-45.0,	35.0,	false);
    fp->player[i++].setValues(8.0,	25.0,	0.5,	0.25,	-30.0,	42.0,	false);
    fp->player[i++].setValues(2.0,	10.0,	0.5,	0.25,	-30.0,	42.0,	false);
    fp->player[i++].setValues(2.0,	-10.0,	0.5,	0.25,	-45.0,	47.0,	false);
    fp->player[i++].setValues(8.0,	-25.0);
    fp->player[i++].setValues(20.0,	10.0);
    fp->player[i++].setValues(20.0,	-10.0);


    fp = &(formations[FT_343_attacking]);
    i = 1;
    fp->player[i++].setValues(-50.0,	0.0,	0.0,	0.0,	0.0,	0.0,	false);
    fp->player[i++].setValues(-16.0,	10.0,	0.1,	0.1,	-50.5,	-30.0,	true);
    fp->player[i++].setValues(-21.0,	0.0,	0.6,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-3.0,	5.0,	0.6,	0.25,	-45.0,	0.0,	true);
    fp->player[i++].setValues(-16.0,	-10.0,	0.6,	0.25,	-45.0,	2.0,	false);
    fp->player[i++].setValues(3.0,	-11.0,	0.6,	0.25,	-30.0,	20.0,	false);
    fp->player[i++].setValues(3.0,	11.0,	0.6,	0.3,	-30.0,	42.0,	false);
    fp->player[i++].setValues(-3.0,	-5.0,	0.5,	0.2,	-45.0,	44.0,	false);
    fp->player[i++].setValues(15.0,	-0.5,	0.6,	0.2,	-2.0,	45.0,	false);
    fp->player[i++].setValues(18.0,	19.0);
    fp->player[i++].setValues(18.0,	-19.0);

}

WorldModel::~WorldModel()
{
}

void
WorldModel::SetupGameStateMap()
{
    mPlayModeMap.clear();

    mPlayModeMap[STR_PM_BeforeKickOff] = PM_BeforeKickOff;
    mPlayModeMap[STR_PM_KickOff_Left]  = PM_KickOff_Left;
    mPlayModeMap[STR_PM_KickOff_Right] = PM_KickOff_Right;
    mPlayModeMap[STR_PM_PlayOn]        = PM_PlayOn;
    mPlayModeMap[STR_PM_KickIn_Left]   = PM_KickIn_Left;
    mPlayModeMap[STR_PM_KickIn_Right]  = PM_KickIn_Right;
    mPlayModeMap[STR_PM_Goal_Left]     = PM_Goal_Left;
    mPlayModeMap[STR_PM_Goal_Right]    = PM_Goal_Right;
    mPlayModeMap[STR_PM_GameOver]      = PM_GameOver;
}

void
WorldModel::SetupVisionObjectMap()
{
    mVisionObjectMap.clear();

    // <ObjectName>+<Id> -> VisionObject
    mVisionObjectMap["Flag1_l"] = VO_FLAG1L;
    mVisionObjectMap["Flag1_r"] = VO_FLAG1R;
    mVisionObjectMap["Flag2_l"] = VO_FLAG2L;
    mVisionObjectMap["Flag2_r"] = VO_FLAG2R;
    mVisionObjectMap["Goal1_l"] = VO_GOAL1L;
    mVisionObjectMap["Goal1_r"] = VO_GOAL1R;
    mVisionObjectMap["Goal2_l"] = VO_GOAL2L;
    mVisionObjectMap["Goal2_r"] = VO_GOAL2R;
    mVisionObjectMap["Ball"]    = VO_BALL;

    mVisionObjectMap["Our1"]    = VO_OUR1;
    mVisionObjectMap["Our2"]    = VO_OUR2;
    mVisionObjectMap["Our3"]    = VO_OUR3;
    mVisionObjectMap["Our4"]    = VO_OUR4;
    mVisionObjectMap["Our5"]    = VO_OUR5;
    mVisionObjectMap["Our6"]    = VO_OUR6;
    mVisionObjectMap["Our7"]    = VO_OUR7;
    mVisionObjectMap["Our8"]    = VO_OUR8;
    mVisionObjectMap["Our9"]    = VO_OUR9;
    mVisionObjectMap["Our10"]    = VO_OUR10;
    mVisionObjectMap["Our11"]    = VO_OUR11;

    mVisionObjectMap["Opp1"]    = VO_OPP1;
    mVisionObjectMap["Opp2"]    = VO_OPP2;
    mVisionObjectMap["Opp3"]    = VO_OPP3;
    mVisionObjectMap["Opp4"]    = VO_OPP4;
    mVisionObjectMap["Opp5"]    = VO_OPP5;
    mVisionObjectMap["Opp6"]    = VO_OPP6;
    mVisionObjectMap["Opp7"]    = VO_OPP7;
    mVisionObjectMap["Opp8"]    = VO_OPP8;
    mVisionObjectMap["Opp9"]    = VO_OPP9;
    mVisionObjectMap["Opp10"]    = VO_OPP10;
    mVisionObjectMap["Opp11"]    = VO_OPP11;
}

bool
WorldModel::ConstructInternal()
{
    mParser = shared_dynamic_cast<BaseParser>(GetCore()->New("SexpParser"));
    if (mParser.get() == 0)
        {
            GetLog()->Error()
                << "ERROR: (WorldModel) create sexpparser parser instance\n.";

            return false;
        }

    SetupGameStateMap();
    SetupVisionObjectMap();

    return true;
}

void
WorldModel::ParseTeamIndex(const Predicate& predicate)
{
    string team;
    if (! predicate.GetValue(predicate.begin(),"team",team))
        {
            return;
        }

    // parse team index
    if (team == "left")
        mTeamIndex = TI_LEFT;
    else if (team == "right")
        mTeamIndex = TI_RIGHT;
    else
        {
            mTeamIndex = TI_NONE;
            GetLog()->Debug()
                << "(WorldModel) received TeamIndex TI_NONE\n";
        }
}

void WorldModel::ParsePlayMode(const Predicate& predicate)
{
    string mode;
    if (! predicate.GetValue(predicate.begin(),"playmode", mode))
        {
            return;
        }

    // parse play mode
    TPlayModeMap::iterator iter = mPlayModeMap.find(mode);
    if (iter != mPlayModeMap.end())
        {
            mPlayMode = (TPlayMode)(*iter).second;
        } else
            {
                GetLog()->Debug()
                    << "ERROR: (WorldModel) could not parse playmode "
                    << mode << "\n";
            }
}

void WorldModel::ParseGameState(const Predicate& predicate)
{
    ParseTeamIndex(predicate);
    ParsePlayMode(predicate);
    predicate.GetValue(predicate.begin(),"FieldLength",   mFieldLength);
    predicate.GetValue(predicate.begin(),"FieldWidth",    mFieldWidth);
    predicate.GetValue(predicate.begin(),"FieldHeigth",   mFieldHeight);
    predicate.GetValue(predicate.begin(),"GoalWidth",     mGoalWidth);
    predicate.GetValue(predicate.begin(),"GoalDepth",     mGoalDepth);
    predicate.GetValue(predicate.begin(),"GoalHeight",    mGoalHeight);
    predicate.GetValue(predicate.begin(),"BorderSize",    mBorderSize);
    predicate.GetValue(predicate.begin(),"time",          mTime);
    predicate.GetValue(predicate.begin(),"unum",          mTeamUnum);
    predicate.GetValue(predicate.begin(),"AgentRadius",   mAgentRadius);
    predicate.GetValue(predicate.begin(),"AgentMass",     mAgentMass);
    predicate.GetValue(predicate.begin(),"AgentMaxSpeed", mAgentMaxSpeed);
    predicate.GetValue(predicate.begin(),"BallRadius",    mBallRadius);
    predicate.GetValue(predicate.begin(),"BallMass",      mBallMass);
}

void WorldModel::ParseObjectVision(const Predicate& predicate)
{
    for (
         Predicate::Iterator iter(predicate);
         iter != iter.end();
         ++iter
         )
        {
            // extract the element as a parameter list
            Predicate::Iterator paramIter = iter;
            if (! predicate.DescentList(paramIter))
                {
                    continue;
                }

            // read the object name
            string name;
            if (! predicate.GetValue(paramIter,name))
            {
                continue;
            }
            if (name == "Player") {
                string strTeam;
                if (predicate.GetValue(paramIter,"team", strTeam)) {
                    if (strTeam == GetTeamName()) {
                        name = "Our";
                    } else {
                        name = "Opp";
                    }
                }
            }
            // try read the 'id' section
            string strId;
            if (predicate.GetValue(paramIter,"id", strId))
                {
                    name += strId;
                }

            // try to lookup the VisionObject
            TVisionObjectMap::iterator iter = mVisionObjectMap.find(name);
            if (iter == mVisionObjectMap.end())
                {
                    GetLog()->Debug() << "cant find in mVisionObjectMap: " << name << endl;
                    continue;
                }

            VisionObject vo = (*iter).second;

            // find  to the 'pol' entry in the object's section
            Predicate::Iterator polIter = paramIter;
            if (! predicate.FindParameter(polIter,"pol"))
                {
                    continue;
                }

            // read the position vector
            VisionSense sense;
            if (
                (! predicate.AdvanceValue(polIter,sense.distance)) ||
                (! predicate.AdvanceValue(polIter,sense.theta)) ||
                (! predicate.AdvanceValue(polIter,sense.phi))
                )
            {
                continue;
            }

            // update the vision map
            mVisionMap[vo] = sense;
        }
    VisionSense myPol;
    myPol.distance = 0;
    myPol.theta = 0;
    myPol.phi = 0;
    mVisionMap[static_cast<VisionObject>(static_cast<int>(VO_OUR0) + GetTeamUnum())] = myPol;
}

void
WorldModel::ParseVision(const Predicate& predicate)
{
    ParseObjectVision(predicate);

    //
    // get our position. The current implementation of the worldmodel
    // relies on the VisionPerceptor that deliverd our position. In
    // future releases this should be replaced by a self localization
    // algorithm using the relative positions of some flags.
    //

    // find the PerfectVision data about the object
    Predicate::Iterator iter(predicate);

    // advance to the section about object 'name'
    if (! predicate.FindParameter(iter,"mypos"))
        {
            mMyPosVEnable = false;
            return;
        }

    // read my position
    VisionSense sense;

    predicate.GetValue(iter,mMyPosV);
    mMyPosVEnable = true;
}

void WorldModel::Parse(const string& message)
{
    shared_ptr<PredicateList> predList =
        mParser->Parse(message);

    if (predList.get() == 0)
    {
            return;
    }

    PredicateList& list = *predList;

    //
    mPositionMap.clear();
    mPositionMap[VO_BALL] = GetObjectPosition(VO_BALL);
    for (VisionObject vo=VO_OUR1; vo <= VO_OUR11; 
         vo = static_cast<VisionObject>(static_cast<int>(vo) + 1)){
        mPositionMap[vo] = GetObjectPosition(vo);
    }
    for (VisionObject vo=VO_OPP1; vo <= VO_OPP11; 
         vo = static_cast<VisionObject>(static_cast<int>(vo) + 1)){
        mPositionMap[vo] = GetObjectPosition(vo);
    }

    //
    for (
         PredicateList::TList::const_iterator iter = list.begin();
         iter != list.end();
         ++iter
         )
        {
            const Predicate& predicate = (*iter);

            if (predicate.name == "GameState")
                {
                    ParseGameState(predicate);
                    continue;
                }

            // check for the Vision perceptor
            if (predicate.name == "Vision")
                {
                    ParseVision(predicate);
                    continue;
                }

            GetLog()->Debug()
                << "(WorldModel) skipping unknown predicate "
                << predicate.name << "\n";
        }

    // update our position
    CalcPlayerPosition();
}

WorldModel::VisionSense WorldModel::GetVisionSense(VisionObject obj)
{
    TVisionMap::iterator iter = mVisionMap.find(obj);

    if (iter == mVisionMap.end())
        {
            GetLog()->Error()
                << "ERROR: (WorldMode::GetVisionSense) unknown VisionObject "
                << obj << "\n";

            return VisionSense();
        }

    return (*iter).second;
}

Vector3f WorldModel::GetPosition(VisionSense sense)
{
    return mMyPos + GetDriveVec(sense) * sense.distance;
}

Vector3f WorldModel::GetObjectPosition(VisionObject obj)
{
    return GetPosition(GetVisionSense(obj));
}

Vector3f WorldModel::GetObjectSpeed(VisionObject obj)
{
    if (GetPlayMode()!=PM_PlayOn ) {
        return Vector3f(0, 0, 0);
    }
    return GetPosition(GetVisionSense(obj)) - mPositionMap[obj];
}

WorldModel::VisionObject WorldModel::GetMyObject()
{
    return static_cast<VisionObject>(static_cast<int>(VO_OUR0) + GetTeamUnum());
}

TPlayMode WorldModel::GetPlayMode()
{
    return mPlayMode;
}

Vector3f WorldModel::GetDriveVec(const WorldModel::VisionSense& vision)
{
    return Vector3f
        (
         gCos(gDegToRad(vision.theta)) *
         gSin(gDegToRad(90.0 - vision.phi)),

         gSin(gDegToRad(vision.theta)) *
         gSin(gDegToRad(90.0 - vision.phi)),

         gCos(gDegToRad(90.0 - vision.phi))
         );
}

void WorldModel::CalcPlayerPosition()
{
    WorldModel::VisionSense fs;
    Vector3f fp;
    Vector3f mp;

    int sign = (GetMyTeam() == TI_RIGHT)? -1: 1;
    //F1L
    fs = GetVisionSense(WorldModel::VO_FLAG1L);
    fp = Vector3f(-GetFieldLength()/2.0,
                  -GetFieldWidth()/2.0,
                  0);
    mp = fp - GetDriveVec(fs) * sign * fs.distance;

    //F2L
    fs = GetVisionSense(WorldModel::VO_FLAG2L);
    fp = Vector3f(-GetFieldLength()/2.0,
                  GetFieldWidth()/2.0,
                  0);
    mp += fp - GetDriveVec(fs) * sign * fs.distance;

    //F1R
    fs = GetVisionSense(WorldModel::VO_FLAG1R);
    fp = Vector3f(GetFieldLength()/2.0,
                  -GetFieldWidth()/2.0,
                  0);
    mp += fp - GetDriveVec(fs) * sign * fs.distance;

    //F2R
    fs = GetVisionSense(WorldModel::VO_FLAG2R);
    fp = Vector3f(GetFieldLength()/2.0,
                  GetFieldWidth()/2.0,
                  0);
    mp += fp - GetDriveVec(fs) * sign * fs.distance;

    mMyPos = mp / 4;

    GetLog()->Debug() << "MYPOS  :" << mMyPos[0] << " " << mMyPos[1] << " " << mMyPos[2] << "\n";
    // $B1&B&$J$iH?E>$5$;$k!#(B
    mMyPos  = mMyPos  * sign;
    mMyPosV = mMyPosV * sign;

    GetLog()->Debug() << "MYPOS  :" << mMyPos[0] << " " << mMyPos[1] << " " << mMyPos[2] << "\n";
    if (mMyPosVEnable) {
        GetLog()->Debug() << "MYPOSV :" << mMyPosV[0] << " " << mMyPosV[1] << " " << mMyPosV[2] << "\n";
        GetLog()->Debug() << "MYPOS error :" << (mMyPosV - mMyPos).Length() << "\n";
    }
}

salt::Vector3f WorldModel::GetMyPosition()
{
    return mMyPos;
}

TTeamIndex WorldModel::GetMyTeam()
{
    return mTeamIndex;
}

float WorldModel::GetMinimalKickDistance()
{
    return mAgentRadius + mBallRadius + 0.04;
}

int WorldModel::GetTeamUnum()
{
    return mTeamUnum;
}


void WorldModel::SetTeamName(std::string name){
    mTeamName =   name;
}

string WorldModel::GetTeamName(){
    return mTeamName;
}

void WorldModel:: SetOppTeamName(std::string name)
{
    mOppTeamName = name;
}

string WorldModel::GetOppTeamName()
{
       return mOppTeamName;    
}


int WorldModel::GetNeedStepsSetTo(VisionObject po, VisionObject vo)
{
    Vector3f pPos = GetObjectPosition(po);
    Vector3f oPos = GetObjectPosition(vo);
    Vector3f oVec = GetObjectSpeed(vo);

    int s;
    for (s = 0; (oPos - pPos).Length() > 0.3 * s && s < 20; s++) {
        oPos += oVec;
    }
    return s;
}

#if 1
WorldModel::VisionObject WorldModel::GetFastestInSetTo(PlayersSet ps, VisionObject vo)
{
    VisionObject objRes;
    VisionObject obj;
    Vector3f objPos = GetObjectPosition(vo);

    double dist;
    double dist2;

    if (ps == SET_TEAMMATE) {
        objRes = VO_OUR1;

        dist = GetNeedStepsSetTo(objRes, vo) * 100 + (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OUR2); i <= static_cast<int>(VO_OUR11); i++) {
            obj = static_cast<VisionObject>(i);
            dist2 = GetNeedStepsSetTo(obj, vo) * 100 + (objPos - GetObjectPosition(obj)).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = obj;
            }
        }
    } 

    if (ps == SET_TEAMMATE_NOGK) {
        objRes = VO_OUR1;

        dist = GetNeedStepsSetTo(objRes, vo) * 100 + (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OUR2); i <= static_cast<int>(VO_OUR11); i++) {
            obj = static_cast<VisionObject>(i);
            dist2 = GetNeedStepsSetTo(obj, vo) * 100 + (objPos - GetObjectPosition(obj)).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = obj;
            }
        }
    } 
    
    if (ps == SET_OPPONENT) {
        objRes = VO_OPP1;
        dist = GetNeedStepsSetTo(objRes, vo) * 100 + (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OPP2); i <= static_cast<int>(VO_OPP2); i++) {
            obj = static_cast<VisionObject>(i);
            dist2 = GetNeedStepsSetTo(obj, vo) * 100 + (objPos - GetObjectPosition(obj)).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = obj;
            }
        }
    } 

    if (ps == SET_ALL) {
        objRes = VO_OUR1;
        dist = GetNeedStepsSetTo(objRes, vo) * 100 + (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OUR2); i <= static_cast<int>(VO_OUR11); i++) {
            obj = static_cast<VisionObject>(i);
            dist2 = GetNeedStepsSetTo(obj, vo) * 100 + (objPos - GetObjectPosition(obj)).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = obj;
            }
        }
        for (int i = static_cast<int>(VO_OPP1); i <= static_cast<int>(VO_OPP11); i++) {
            obj = static_cast<VisionObject>(i);
            dist2 = GetNeedStepsSetTo(obj, vo) * 100 + (objPos - GetObjectPosition(obj)).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = obj;
            }
        }
    } 
    
    return objRes;
}

#else
WorldModel::VisionObject WorldModel::GetFastestInSetTo(PlayersSet ps, VisionObject vo)
{
    salt::Vector3f objPos = GetObjectPosition(vo);
    VisionObject objRes;
    double dist;
    double dist2;

    if (ps == SET_TEAMMATE) {
        objRes = VO_OUR1;
        dist = (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OUR2); i <= static_cast<int>(VO_OUR11); i++) {
            dist2 = (objPos - GetObjectPosition(static_cast<VisionObject>(i))).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = static_cast<VisionObject>(i);
            }
        }
    } 

    if (ps == SET_TEAMMATE_NOGK) {
        objRes = VO_OUR2;
        dist = (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OUR3); i <= static_cast<int>(VO_OUR11); i++) {
            dist2 = (objPos - GetObjectPosition(static_cast<VisionObject>(i))).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = static_cast<VisionObject>(i);
            }
        }
    } 
    
    if (ps == SET_OPPONENT) {
        objRes = VO_OPP1;
        dist = (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OPP2); i <= static_cast<int>(VO_OPP2); i++) {
            dist2 = (objPos - GetObjectPosition(static_cast<VisionObject>(i))).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = static_cast<VisionObject>(i);
            }
        }
    } 

    if (ps == SET_ALL) {
        objRes = VO_OUR1;
        dist = (objPos - GetObjectPosition(objRes)).Length();

        for (int i = static_cast<int>(VO_OUR2); i <= static_cast<int>(VO_OUR11); i++) {
            dist2 = (objPos - GetObjectPosition(static_cast<VisionObject>(i))).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = static_cast<VisionObject>(i);
            }
        }
        for (int i = static_cast<int>(VO_OPP1); i <= static_cast<int>(VO_OPP11); i++) {
            dist2 = (objPos - GetObjectPosition(static_cast<VisionObject>(i))).Length();
            if (dist > dist2) {
                dist = dist2;
                objRes = static_cast<VisionObject>(i);
            }
        }
    } 
    
    return objRes;
}
#endif

salt::Vector3f WorldModel::GetStrategicPosition()
{
    double x, y;
    // get the home position and calculate the associated strategic position
    //PlayerData& myData = formations[FT_433_offensive].player[GetTeamUnum()];
    PlayerData& myData = formations[FT_334_offensive].player[GetTeamUnum()];
    Vector3f ballPos = GetObjectPosition(VO_BALL);

    y = myData.pos.y() + ballPos.y() * myData.attr.y();
    x = myData.pos.x() + ballPos.x() * myData.attr.x();

    // do not move too much to the side
    /*
    if( fabs( y ) > 0.5*dMaxYPercentage*PITCH_WIDTH )
        y = sign(y)*0.5*dMaxYPercentage*PITCH_WIDTH;
    */

    // when behind ball is set, do not move to point in front of ball
    if( myData.behindball == true && x > ballPos.x() ) {
        x = ballPos.x();
    }

    // do not move past maximal x or before minimal x
    if( x > myData.maxx ) {
        x = myData.maxx;
    } else if( x < myData.minx ) {
        x = myData.minx;
    }

    // when x coordinate is in front of allowed x value, change it
    /*
    if( x > dMaxXInPlayMode )
        x = dMaxXInPlayMode;
    */
  
    return salt::Vector3f(x, y, 0);
}

salt::Vector3f WorldModel::GetFromMyPosTo(salt::Vector3f pos)
{
    return (mTeamIndex == TI_LEFT)?
        pos - mMyPos:
        pos + mMyPos;
}
